# This reads the file specified as the first argument
# and creates a DF3 file appropriate to POV-Ray.
# Copyright 2009 Darren New
# Released under the WTFPL license
# See http://en.wikipedia.org/wiki/WTFPL for details

proc main {} {
    global argv
    if {1 != [llength $argv] || ![file readable [lindex $argv 0]]} {
	help
	exit 1
    }
    set infname [lindex $argv 0]
    set outfname [file rootname $infname].df3
    set inf [open $infname]
    set parses [scan_file $inf]
    puts "X,Y,Z,bytes per voxel: $parses"
    set outf [open $outfname w]
    generate_output $inf $parses $outf
    close $inf
    close $outf
    puts "Success!"
    exit 0
}

proc help {} {
    catch {console show}
    puts "This program reads a file full of data for a DF3 file"
    puts "(a la POV-Ray ray tracer) and creates the appropriate"
    puts "binary file. The output file has the same name as the"
    puts "input file with .df3 replacing the extension (or with"
    puts ".df3 appended if there is no extension). The input is"
    puts "a file with a collection of 4-tuples. Each 4-tuple is"
    puts "integers (in ASCII) representing x, y, z, and a value"
    puts "to be inserted into the DF3. There is no need for any"
    puts "header, as ranges are taken from the file itself. You"
    puts "must put four integers per line in this version. This"
    puts "can be slow because it assumes a very big input file."
    after 3000
    puts "Copyright 2009 Darren New"
    puts "Released under the WTFPL license"
    puts "See http://en.wikipedia.org/wiki/WTFPL for details"
}

# Scans thru the input file, returns maximum of x, y, and z,
# and also returns the number of bytes each voxel will take.
#  (Yes, the loop here could be factored out.)
proc scan_file {inf} {
    set maxx -1; set maxy -1; set maxz -1; set maxv -1
    seek $inf 0
    set line [gets $inf]
    while {![eof $inf]} {
	set c [scan $line "%d %d %d %d" x y z v]
	if {$c != 4} {puts "Bad line: $line"}
	if {$maxx < $x} {set maxx $x}
	if {$maxy < $y} {set maxy $y}
	if {$maxz < $z} {set maxz $z}
	if {$maxv < $v} {set maxv $v}
	set line [gets $inf]
    }
    if {$maxv < 256} {
	set bytes 1 
    } elseif {$maxv < 65536} {
	set bytes 2
    } else {
	set bytes 4
    }
    incr maxx; incr maxy; incr maxz ; # Make it open interval.

    # Not sure how this'll behave on all machines, so skip it
    if {0} {
	set maxsize [expr {2147483648 / $bytes}]
	if {$maxsize <= ($x * $y * $z)} {
	    error "File result too large"
	}
    }
    return [list $maxx $maxy $maxz $bytes]
}

proc generate_output {inf parses outf} {
    seek $inf 0
    fconfigure $outf -encoding binary -translation binary
    set maxx [lindex $parses 0]
    set maxy [lindex $parses 1]
    set maxz [lindex $parses 2]
    set bytes [lindex $parses 3]

    # Figure out the byte format to use for data
    if {$bytes == 1} {set fmt c}
    if {$bytes == 2} {set fmt S}
    if {$bytes == 4} {set fmt I}

    # Generate the header
    set head [binary format S3 [list $maxx $maxy $maxz]]
    puts -nonewline $outf $head

    # Fill in the empty file, so defaults are 0.
    for {set x 0} {$x < $maxx} {incr x} {
	for {set y 0} {$y < $maxy} {incr y} {
	    for {set z 0} {$z < $maxz} {incr z} {
		puts -nonewline $outf [binary format $fmt 0]
	    }
	}
    }

    # Now read the input file and fill in selected words.
    set line [gets $inf]
    set count 0
    while {![eof $inf]} {
	set c [scan $line "%d %d %d %d" x y z v]
	set offs [calc_offs $x $y $z $maxx $maxy $maxz $bytes]
	seek $outf $offs
	puts -nonewline $outf [binary format $fmt $v]
	incr count
	set line [gets $inf]
    }
    puts "Wrote $count voxels to file"
}

# Return the offset in the file where the bytes should be written.
proc calc_offs {x y z maxx maxy maxz bytes} {
    set inx [expr {$x + ($maxx * $y) + ($maxx * $maxy * $z)}]
    set offs [expr {($bytes * $inx) + 6}] ; # 6 bytes in header
    return $offs
}

main


